from __future__ import annotations

from typing import Any

from hatchling.builders.hooks.plugin.interface import BuildHookInterface
from hatchling.version.core import VersionFile


class VersionBuildHook(BuildHookInterface):
    PLUGIN_NAME = 'version'

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self.__config_path: str | None = None
        self.__config_template: str | None = None
        self.__config_pattern: str | bool | None = None

    @property
    def config_path(self) -> str:
        if self.__config_path is None:
            path = self.config.get('path', '')
            if not isinstance(path, str):
                message = f'Option `path` for build hook `{self.PLUGIN_NAME}` must be a string'
                raise TypeError(message)

            if not path:
                message = f'Option `path` for build hook `{self.PLUGIN_NAME}` is required'
                raise ValueError(message)

            self.__config_path = path

        return self.__config_path

    @property
    def config_template(self) -> str:
        if self.__config_template is None:
            template = self.config.get('template', '')
            if not isinstance(template, str):
                message = f'Option `template` for build hook `{self.PLUGIN_NAME}` must be a string'
                raise TypeError(message)

            self.__config_template = template

        return self.__config_template

    @property
    def config_pattern(self) -> str | bool:
        if self.__config_pattern is None:
            pattern = self.config.get('pattern', '')
            if not isinstance(pattern, (str, bool)):
                message = f'Option `pattern` for build hook `{self.PLUGIN_NAME}` must be a string or a boolean'
                raise TypeError(message)

            self.__config_pattern = pattern

        return self.__config_pattern

    def initialize(
        self,
        version: str,  # noqa: ARG002
        build_data: dict[str, Any],
    ) -> None:
        version_file = VersionFile(self.root, self.config_path)
        if self.config_pattern:
            version_file.read(pattern=self.config_pattern)
            version_file.set_version(self.metadata.version)
        else:
            version_file.write(self.metadata.version, self.config_template)

        build_data['artifacts'].append(f'/{self.config_path}')



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable

if TYPE_CHECKING:
    from hatchling.metadata.core import ProjectMetadata

DEFAULT_METADATA_VERSION = '2.4'
LATEST_METADATA_VERSION = '2.4'
CORE_METADATA_PROJECT_FIELDS = {
    'Author': ('authors',),
    'Author-email': ('authors',),
    'Classifier': ('classifiers',),
    'Description': ('readme',),
    'Description-Content-Type': ('readme',),
    'Dynamic': ('dynamic',),
    'Keywords': ('keywords',),
    'License': ('license',),
    'License-Expression': ('license',),
    'License-Files': ('license-files',),
    'Maintainer': ('maintainers',),
    'Maintainer-email': ('maintainers',),
    'Name': ('name',),
    'Provides-Extra': ('dependencies', 'optional-dependencies'),
    'Requires-Dist': ('dependencies',),
    'Requires-Python': ('requires-python',),
    'Summary': ('description',),
    'Project-URL': ('urls',),
    'Version': ('version',),
}
PROJECT_CORE_METADATA_FIELDS = {
    'authors': ('Author', 'Author-email'),
    'classifiers': ('Classifier',),
    'dependencies': ('Requires-Dist',),
    'dynamic': ('Dynamic',),
    'keywords': ('Keywords',),
    'license': ('License', 'License-Expression'),
    'license-files': ('License-Files',),
    'maintainers': ('Maintainer', 'Maintainer-email'),
    'name': ('Name',),
    'optional-dependencies': ('Requires-Dist', 'Provides-Extra'),
    'readme': ('Description', 'Description-Content-Type'),
    'requires-python': ('Requires-Python',),
    'description': ('Summary',),
    'urls': ('Project-URL',),
    'version': ('Version',),
}


def get_core_metadata_constructors() -> dict[str, Callable]:
    """
    https://packaging.python.org/specifications/core-metadata/
    """
    return {
        '1.2': construct_metadata_file_1_2,
        '2.1': construct_metadata_file_2_1,
        '2.2': construct_metadata_file_2_2,
        '2.3': construct_metadata_file_2_3,
        '2.4': construct_metadata_file_2_4,
    }


def project_metadata_from_core_metadata(core_metadata: str) -> dict[str, Any]:
    # https://packaging.python.org/en/latest/specifications/core-metadata/
    import email
    from email.headerregistry import HeaderRegistry

    header_registry = HeaderRegistry()

    message = email.message_from_string(core_metadata)
    metadata: dict[str, Any] = {}

    if name := message.get('Name'):
        metadata['name'] = name
    else:
        error_message = 'Missing required core metadata: Name'
        raise ValueError(error_message)

    if version := message.get('Version'):
        metadata['version'] = version
    else:
        error_message = 'Missing required core metadata: Version'
        raise ValueError(error_message)

    if (dynamic_fields := message.get_all('Dynamic')) is not None:
        # Use as an ordered set to retain bidirectional formatting.
        # This likely doesn't matter but we try hard around here.
        metadata['dynamic'] = list({
            project_field: None
            for core_metadata_field in dynamic_fields
            for project_field in CORE_METADATA_PROJECT_FIELDS.get(core_metadata_field, ())
        })

    if description := message.get_payload():
        metadata['readme'] = {
            'content-type': message.get('Description-Content-Type', 'text/plain'),
            'text': description,
        }

    if (license_expression := message.get('License-Expression')) is not None:
        metadata['license'] = license_expression
    elif (license_text := message.get('License')) is not None:
        metadata['license'] = {'text': license_text}

    if (license_files := message.get_all('License-File')) is not None:
        metadata['license-files'] = license_files

    if (summary := message.get('Summary')) is not None:
        metadata['description'] = summary

    if (keywords := message.get('Keywords')) is not None:
        metadata['keywords'] = keywords.split(',')

    if (classifiers := message.get_all('Classifier')) is not None:
        metadata['classifiers'] = classifiers

    if (project_urls := message.get_all('Project-URL')) is not None:
        urls = {}
        for project_url in project_urls:
            label, url = project_url.split(',', maxsplit=1)
            urls[label.strip()] = url.strip()
        metadata['urls'] = urls

    authors = []
    if (author := message.get('Author')) is not None:
        authors.append({'name': author})

    if (author_email := message.get('Author-email')) is not None:
        address_header = header_registry('resent-from', author_email)
        for address in address_header.addresses:  # type: ignore[attr-defined]
            data = {'email': address.addr_spec}
            if name := address.display_name:
                data['name'] = name
            authors.append(data)

    if authors:
        metadata['authors'] = authors

    maintainers = []
    if (maintainer := message.get('Maintainer')) is not None:
        maintainers.append({'name': maintainer})

    if (maintainer_email := message.get('Maintainer-email')) is not None:
        address_header = header_registry('resent-from', maintainer_email)
        for address in address_header.addresses:  # type: ignore[attr-defined]
            data = {'email': address.addr_spec}
            if name := address.display_name:
                data['name'] = name
            maintainers.append(data)

    if maintainers:
        metadata['maintainers'] = maintainers

    if (requires_python := message.get('Requires-Python')) is not None:
        metadata['requires-python'] = requires_python

    optional_dependencies: dict[str, list[str]] = {}
    if (extras := message.get_all('Provides-Extra')) is not None:
        for extra in extras:
            optional_dependencies[extra] = []

    if (requirements := message.get_all('Requires-Dist')) is not None:
        from packaging.requirements import Requirement

        dependencies = []
        for requirement in requirements:
            req = Requirement(requirement)
            if req.marker is None:
                dependencies.append(str(req))
                continue

            markers = req.marker._markers  # noqa: SLF001
            for i, marker in enumerate(markers):
                if isinstance(marker, tuple):
                    left, _, right = marker
                    if left.value == 'extra':
                        extra = right.value
                        del markers[i]  # noqa: B909
                        # If there was only one marker then there will be an unnecessary
                        # trailing semicolon in the string representation
                        if not markers:
                            req.marker = None
                        # Otherwise we need to remove the preceding `and` operation
                        else:
                            del markers[i - 1]

                        optional_dependencies.setdefault(extra, []).append(str(req))
                        break
            else:
                dependencies.append(str(req))

        metadata['dependencies'] = dependencies

    if optional_dependencies:
        metadata['optional-dependencies'] = optional_dependencies

    return metadata


def construct_metadata_file_1_2(metadata: ProjectMetadata, extra_dependencies: tuple[str] | None = None) -> str:
    """
    https://peps.python.org/pep-0345/
    """
    metadata_file = 'Metadata-Version: 1.2\n'
    metadata_file += f'Name: {metadata.core.raw_name}\n'
    metadata_file += f'Version: {metadata.version}\n'

    if metadata.core.description:
        metadata_file += f'Summary: {metadata.core.description}\n'

    if metadata.core.urls:
        for label, url in metadata.core.urls.items():
            metadata_file += f'Project-URL: {label}, {url}\n'

    authors_data = metadata.core.authors_data
    if authors_data['name']:
        metadata_file += f"Author: {', '.join(authors_data['name'])}\n"
    if authors_data['email']:
        metadata_file += f"Author-email: {', '.join(authors_data['email'])}\n"

    maintainers_data = metadata.core.maintainers_data
    if maintainers_data['name']:
        metadata_file += f"Maintainer: {', '.join(maintainers_data['name'])}\n"
    if maintainers_data['email']:
        metadata_file += f"Maintainer-email: {', '.join(maintainers_data['email'])}\n"

    if metadata.core.license:
        license_start = 'License: '
        indent = ' ' * (len(license_start) - 1)
        metadata_file += license_start

        for i, line in enumerate(metadata.core.license.splitlines()):
            if i == 0:
                metadata_file += f'{line}\n'
            else:
                metadata_file += f'{indent}{line}\n'
    elif metadata.core.license_expression:
        metadata_file += f'License: {metadata.core.license_expression}\n'

    if metadata.core.keywords:
        metadata_file += f"Keywords: {','.join(metadata.core.keywords)}\n"

    if metadata.core.classifiers:
        for classifier in metadata.core.classifiers:
            metadata_file += f'Classifier: {classifier}\n'

    if metadata.core.requires_python:
        metadata_file += f'Requires-Python: {metadata.core.requires_python}\n'

    if metadata.core.dependencies:
        for dependency in metadata.core.dependencies:
            metadata_file += f'Requires-Dist: {dependency}\n'

    if extra_dependencies:
        for dependency in extra_dependencies:
            metadata_file += f'Requires-Dist: {dependency}\n'

    return metadata_file


def construct_metadata_file_2_1(metadata: ProjectMetadata, extra_dependencies: tuple[str] | None = None) -> str:
    """
    https://peps.python.org/pep-0566/
    """
    metadata_file = 'Metadata-Version: 2.1\n'
    metadata_file += f'Name: {metadata.core.raw_name}\n'
    metadata_file += f'Version: {metadata.version}\n'

    if metadata.core.description:
        metadata_file += f'Summary: {metadata.core.description}\n'

    if metadata.core.urls:
        for label, url in metadata.core.urls.items():
            metadata_file += f'Project-URL: {label}, {url}\n'

    authors_data = metadata.core.authors_data
    if authors_data['name']:
        metadata_file += f"Author: {', '.join(authors_data['name'])}\n"
    if authors_data['email']:
        metadata_file += f"Author-email: {', '.join(authors_data['email'])}\n"

    maintainers_data = metadata.core.maintainers_data
    if maintainers_data['name']:
        metadata_file += f"Maintainer: {', '.join(maintainers_data['name'])}\n"
    if maintainers_data['email']:
        metadata_file += f"Maintainer-email: {', '.join(maintainers_data['email'])}\n"

    if metadata.core.license:
        license_start = 'License: '
        indent = ' ' * (len(license_start) - 1)
        metadata_file += license_start

        for i, line in enumerate(metadata.core.license.splitlines()):
            if i == 0:
                metadata_file += f'{line}\n'
            else:
                metadata_file += f'{indent}{line}\n'
    elif metadata.core.license_expression:
        metadata_file += f'License: {metadata.core.license_expression}\n'

    if metadata.core.keywords:
        metadata_file += f"Keywords: {','.join(metadata.core.keywords)}\n"

    if metadata.core.classifiers:
        for classifier in metadata.core.classifiers:
            metadata_file += f'Classifier: {classifier}\n'

    if metadata.core.requires_python:
        metadata_file += f'Requires-Python: {metadata.core.requires_python}\n'

    if metadata.core.dependencies:
        for dependency in metadata.core.dependencies:
            metadata_file += f'Requires-Dist: {dependency}\n'

    if extra_dependencies:
        for dependency in extra_dependencies:
            metadata_file += f'Requires-Dist: {dependency}\n'

    if metadata.core.optional_dependencies:
        for option, dependencies in metadata.core.optional_dependencies.items():
            metadata_file += f'Provides-Extra: {option}\n'
            for dependency in dependencies:
                if ';' in dependency:
                    dep_name, dep_env_marker = dependency.split(';', maxsplit=1)
                    metadata_file += f'Requires-Dist: {dep_name}; ({dep_env_marker.strip()}) and extra == {option!r}\n'
                elif '@ ' in dependency:
                    metadata_file += f'Requires-Dist: {dependency} ; extra == {option!r}\n'
                else:
                    metadata_file += f'Requires-Dist: {dependency}; extra == {option!r}\n'

    if metadata.core.readme:
        metadata_file += f'Description-Content-Type: {metadata.core.readme_content_type}\n'
        metadata_file += f'\n{metadata.core.readme}'

    return metadata_file


def construct_metadata_file_2_2(metadata: ProjectMetadata, extra_dependencies: tuple[str] | None = None) -> str:
    """
    https://peps.python.org/pep-0643/
    """
    metadata_file = 'Metadata-Version: 2.2\n'
    metadata_file += f'Name: {metadata.core.raw_name}\n'
    metadata_file += f'Version: {metadata.version}\n'

    if metadata.core.dynamic:
        # Ordered set
        for field in {
            core_metadata_field: None
            for project_field in metadata.core.dynamic
            for core_metadata_field in PROJECT_CORE_METADATA_FIELDS.get(project_field, ())
        }:
            metadata_file += f'Dynamic: {field}\n'

    if metadata.core.description:
        metadata_file += f'Summary: {metadata.core.description}\n'

    if metadata.core.urls:
        for label, url in metadata.core.urls.items():
            metadata_file += f'Project-URL: {label}, {url}\n'

    authors_data = metadata.core.authors_data
    if authors_data['name']:
        metadata_file += f"Author: {', '.join(authors_data['name'])}\n"
    if authors_data['email']:
        metadata_file += f"Author-email: {', '.join(authors_data['email'])}\n"

    maintainers_data = metadata.core.maintainers_data
    if maintainers_data['name']:
        metadata_file += f"Maintainer: {', '.join(maintainers_data['name'])}\n"
    if maintainers_data['email']:
        metadata_file += f"Maintainer-email: {', '.join(maintainers_data['email'])}\n"

    if metadata.core.license:
        license_start = 'License: '
        indent = ' ' * (len(license_start) - 1)
        metadata_file += license_start

        for i, line in enumerate(metadata.core.license.splitlines()):
            if i == 0:
                metadata_file += f'{line}\n'
            else:
                metadata_file += f'{indent}{line}\n'
    elif metadata.core.license_expression:
        metadata_file += f'License: {metadata.core.license_expression}\n'

    if metadata.core.keywords:
        metadata_file += f"Keywords: {','.join(metadata.core.keywords)}\n"

    if metadata.core.classifiers:
        for classifier in metadata.core.classifiers:
            metadata_file += f'Classifier: {classifier}\n'

    if metadata.core.requires_python:
        metadata_file += f'Requires-Python: {metadata.core.requires_python}\n'

    if metadata.core.dependencies:
        for dependency in metadata.core.dependencies:
            metadata_file += f'Requires-Dist: {dependency}\n'

    if extra_dependencies:
        for dependency in extra_dependencies:
            metadata_file += f'Requires-Dist: {dependency}\n'

    if metadata.core.optional_dependencies:
        for option, dependencies in metadata.core.optional_dependencies.items():
            metadata_file += f'Provides-Extra: {option}\n'
            for dependency in dependencies:
                if ';' in dependency:
                    dep_name, dep_env_marker = dependency.split(';', maxsplit=1)
                    metadata_file += f'Requires-Dist: {dep_name}; ({dep_env_marker.strip()}) and extra == {option!r}\n'
                elif '@ ' in dependency:
                    metadata_file += f'Requires-Dist: {dependency} ; extra == {option!r}\n'
                else:
                    metadata_file += f'Requires-Dist: {dependency}; extra == {option!r}\n'

    if metadata.core.readme:
        metadata_file += f'Description-Content-Type: {metadata.core.readme_content_type}\n'
        metadata_file += f'\n{metadata.core.readme}'

    return metadata_file


def construct_metadata_file_2_3(metadata: ProjectMetadata, extra_dependencies: tuple[str] | None = None) -> str:
    """
    https://peps.python.org/pep-0685/
    """
    metadata_file = 'Metadata-Version: 2.3\n'
    metadata_file += f'Name: {metadata.core.raw_name}\n'
    metadata_file += f'Version: {metadata.version}\n'

    if metadata.core.dynamic:
        # Ordered set
        for field in {
            core_metadata_field: None
            for project_field in metadata.core.dynamic
            for core_metadata_field in PROJECT_CORE_METADATA_FIELDS.get(project_field, ())
        }:
            metadata_file += f'Dynamic: {field}\n'

    if metadata.core.description:
        metadata_file += f'Summary: {metadata.core.description}\n'

    if metadata.core.urls:
        for label, url in metadata.core.urls.items():
            metadata_file += f'Project-URL: {label}, {url}\n'

    authors_data = metadata.core.authors_data
    if authors_data['name']:
        metadata_file += f"Author: {', '.join(authors_data['name'])}\n"
    if authors_data['email']:
        metadata_file += f"Author-email: {', '.join(authors_data['email'])}\n"

    maintainers_data = metadata.core.maintainers_data
    if maintainers_data['name']:
        metadata_file += f"Maintainer: {', '.join(maintainers_data['name'])}\n"
    if maintainers_data['email']:
        metadata_file += f"Maintainer-email: {', '.join(maintainers_data['email'])}\n"

    if metadata.core.license:
        license_start = 'License: '
        indent = ' ' * (len(license_start) - 1)
        metadata_file += license_start

        for i, line in enumerate(metadata.core.license.splitlines()):
            if i == 0:
                metadata_file += f'{line}\n'
            else:
                metadata_file += f'{indent}{line}\n'
    elif metadata.core.license_expression:
        metadata_file += f'License: {metadata.core.license_expression}\n'

    if metadata.core.keywords:
        metadata_file += f"Keywords: {','.join(metadata.core.keywords)}\n"

    if metadata.core.classifiers:
        for classifier in metadata.core.classifiers:
            metadata_file += f'Classifier: {classifier}\n'

    if metadata.core.requires_python:
        metadata_file += f'Requires-Python: {metadata.core.requires_python}\n'

    if metadata.core.dependencies:
        for dependency in metadata.core.dependencies:
            metadata_file += f'Requires-Dist: {dependency}\n'

    if extra_dependencies:
        for dependency in extra_dependencies:
            metadata_file += f'Requires-Dist: {dependency}\n'

    if metadata.core.optional_dependencies:
        for option, dependencies in metadata.core.optional_dependencies.items():
            metadata_file += f'Provides-Extra: {option}\n'
            for dependency in dependencies:
                if ';' in dependency:
                    dep_name, dep_env_marker = dependency.split(';', maxsplit=1)
                    metadata_file += f'Requires-Dist: {dep_name}; ({dep_env_marker.strip()}) and extra == {option!r}\n'
                elif '@ ' in dependency:
                    metadata_file += f'Requires-Dist: {dependency} ; extra == {option!r}\n'
                else:
                    metadata_file += f'Requires-Dist: {dependency}; extra == {option!r}\n'

    if metadata.core.readme:
        metadata_file += f'Description-Content-Type: {metadata.core.readme_content_type}\n'
        metadata_file += f'\n{metadata.core.readme}'

    return metadata_file


def construct_metadata_file_2_4(metadata: ProjectMetadata, extra_dependencies: tuple[str] | None = None) -> str:
    """
    https://peps.python.org/pep-0639/
    """
    metadata_file = 'Metadata-Version: 2.4\n'
    metadata_file += f'Name: {metadata.core.raw_name}\n'
    metadata_file += f'Version: {metadata.version}\n'

    if metadata.core.dynamic:
        # Ordered set
        for field in {
            core_metadata_field: None
            for project_field in metadata.core.dynamic
            for core_metadata_field in PROJECT_CORE_METADATA_FIELDS.get(project_field, ())
        }:
            metadata_file += f'Dynamic: {field}\n'

    if metadata.core.description:
        metadata_file += f'Summary: {metadata.core.description}\n'

    if metadata.core.urls:
        for label, url in metadata.core.urls.items():
            metadata_file += f'Project-URL: {label}, {url}\n'

    authors_data = metadata.core.authors_data
    if authors_data['name']:
        metadata_file += f"Author: {', '.join(authors_data['name'])}\n"
    if authors_data['email']:
        metadata_file += f"Author-email: {', '.join(authors_data['email'])}\n"

    maintainers_data = metadata.core.maintainers_data
    if maintainers_data['name']:
        metadata_file += f"Maintainer: {', '.join(maintainers_data['name'])}\n"
    if maintainers_data['email']:
        metadata_file += f"Maintainer-email: {', '.join(maintainers_data['email'])}\n"

    if metadata.core.license:
        license_start = 'License: '
        indent = ' ' * (len(license_start) - 1)
        metadata_file += license_start

        for i, line in enumerate(metadata.core.license.splitlines()):
            if i == 0:
                metadata_file += f'{line}\n'
            else:
                metadata_file += f'{indent}{line}\n'

    if metadata.core.license_expression:
        metadata_file += f'License-Expression: {metadata.core.license_expression}\n'

    if metadata.core.license_files:
        for license_file in metadata.core.license_files:
            metadata_file += f'License-File: {license_file}\n'

    if metadata.core.keywords:
        metadata_file += f"Keywords: {','.join(metadata.core.keywords)}\n"

    if metadata.core.classifiers:
        for classifier in metadata.core.classifiers:
            metadata_file += f'Classifier: {classifier}\n'

    if metadata.core.requires_python:
        metadata_file += f'Requires-Python: {metadata.core.requires_python}\n'

    if metadata.core.dependencies:
        for dependency in metadata.core.dependencies:
            metadata_file += f'Requires-Dist: {dependency}\n'

    if extra_dependencies:
        for dependency in extra_dependencies:
            metadata_file += f'Requires-Dist: {dependency}\n'

    if metadata.core.optional_dependencies:
        for option, dependencies in metadata.core.optional_dependencies.items():
            metadata_file += f'Provides-Extra: {option}\n'
            for dependency in dependencies:
                if ';' in dependency:
                    dep_name, dep_env_marker = dependency.split(';', maxsplit=1)
                    metadata_file += f'Requires-Dist: {dep_name}; ({dep_env_marker.strip()}) and extra == {option!r}\n'
                elif '@ ' in dependency:
                    metadata_file += f'Requires-Dist: {dependency} ; extra == {option!r}\n'
                else:
                    metadata_file += f'Requires-Dist: {dependency}; extra == {option!r}\n'

    if metadata.core.readme:
        metadata_file += f'Description-Content-Type: {metadata.core.readme_content_type}\n'
        metadata_file += f'\n{metadata.core.readme}'

    return metadata_file



import logging
import math
import os
from io import BytesIO
from pathlib import Path

import requests
from fitrequest.client import FitRequest

from skillcorner_on_demand.method import METHOD_DOCSTRING, METHODS_BINDING

logger = logging.getLogger(__name__)

BASE_URL = 'https://tactical.skillcorner.com'
BASE_CLIENT_NAME = 'skillcorner_on_demand'
USERNAME = os.getenv('SKILLCORNER_ON_DEMAND_USERNAME')
PASSWORD = os.getenv('SKILLCORNER_ON_DEMAND_PASSWORD')


class PreservingFile:
    def __init__(self, file_object):
        self.file_object = file_object
        self._content = None

    def read(self, *args, **kwargs):
        if self._content is None:
            current_pos = self.file_object.tell()
            self.file_object.seek(0)
            self._content = self.file_object.read()
            self.file_object.seek(current_pos)
        return BytesIO(self._content).read(*args, **kwargs)


class SkillcornerOnDemandClient(FitRequest):
    """Client for the Skillcorner On Demand API."""

    base_url = BASE_URL
    base_client_name = BASE_CLIENT_NAME
    _docstring_template = METHOD_DOCSTRING
    _methods_binding = METHODS_BINDING

    def _check_request(self, method, url, **kwargs):
        kwargs_copy = kwargs.copy()
        if 'files' in kwargs_copy:
            preserving_files = {
                k: (v[0], PreservingFile(v[1]), v[2])
                for k, v in kwargs_copy['files'].items()
            }
            kwargs_copy['files'] = preserving_files
        super()._check_request(method, url, **kwargs_copy)

    def upload_match_video(
        self, match_id: int, path: str, is_update: bool = False
    ) -> dict:
        """Upload a video file to the server.

        Args:
            match_id (int): The match id.
            video_path (string): The path to the video file.
            is_update (bool): Whether to update the video file.

        Returns:
            string: The response from the server.

        Examples:
        ```py
            client.upload_match_video(1, 'path/to/my/video.mp4')
            # {'status': 201}
        ```

        """
        try:
            video_parts = self._get_number_of_parts(path)
            presigned_urls, upload_id = self._request_presigned_urls(
                match_id, video_parts, is_update
            )
            parts = self._multipart_upload(presigned_urls, path)
            return self.complete_multipart_upload(match_id, parts, upload_id, is_update)
        except Exception as e:
            return e

    def _request_presigned_urls(
        self, match_id: int, parts: int, is_update: bool = False
    ) -> tuple[list[str], str]:
        if is_update:
            res = self.session.put(
                self.base_url + f'/api/request/{match_id}/video/',
                json={'parts': parts},
                headers=self.session.headers,
            )
        else:
            res = self.session.post(
                self.base_url + f'/api/request/{match_id}/video/',
                json={'parts': parts},
                headers=self.session.headers,
            )
        return res.json()['presignedUrls'], res.json()['uploadId']

    def _multipart_upload(
        self, presigned_urls: list[str], video_path: Path
    ) -> list[dict]:
        part_size = 50 * 1024 * 1024
        parts = []
        print('Uploading video to the server...')
        with open(video_path, 'rb') as video_file:
            for i, presigned_url in enumerate(presigned_urls):
                video_chunk = video_file.read(part_size)
                etag = self._upload_part(presigned_url, video_chunk)
                parts.append({'PartNumber': i + 1, 'ETag': etag})
                print(f'Uploaded part {i + 1} / {len(presigned_urls)}')
        return parts

    def _upload_part(self, presigned_url: str, video_chunk) -> str:
        res = requests.put(presigned_url, data=video_chunk)
        res.raise_for_status()
        etag = res.headers['ETag']
        if not etag:
            raise ValueError(f'ETag not found in response headers: {res.headers}')
        return etag

    def complete_multipart_upload(
        self,
        match_id: int,
        parts: list[dict],
        upload_id: int,
        is_update: bool = False,
    ) -> requests.Response:
        if is_update:
            res = self.session.put(
                self.base_url + f'/api/request/{match_id}/video/',
                json={'parts': parts, 'complete': True, 'uploadId': upload_id},
                headers=self.session.headers,
            )
        else:
            res = self.session.post(
                self.base_url + f'/api/request/{match_id}/video/',
                json={'parts': parts, 'complete': True, 'uploadId': upload_id},
                headers=self.session.headers,
            )
        return res

    def _get_number_of_parts(self, video_path):
        file_size = os.path.getsize(video_path)
        return math.ceil(file_size / (50 * 1024 * 1024))

    def upload_match_sheet(self, match_id: int, path: str, is_update: bool = False):
        """Post a match sheet to the server.

        Args:
            match_id (int): The match id.
            match_sheet (dict): Path of the file.
            is_update (bool): Whether to update the match sheet.

        Returns:
            dict: The response from the server.

        Examples:
        ```py
            response = client.post_match_sheet(1, 'path/to/my/file.csv')
            # {"message": "Information successfully added to the request"}
        ```

        """
        with open(path, 'rb') as match_sheet_file:
            try:
                if is_update:
                    return self.update_match_sheet(
                        match_id,
                        files={
                            'file': (
                                'match_sheet.csv',
                                match_sheet_file,
                                'text/csv',
                            )
                        },
                    )
            except Exception as e:
                return e
            try:
                return self.post_match_sheet(
                    match_id,
                    files={'file': ('match_sheet.csv', match_sheet_file, 'text/csv')},
                )
            except Exception as e:
                return e

    def upload_period_limits(self, match_id: int, path: str, is_update: bool = False):
        """Post period limits to the server.

        Args:
            match_id (int): The match id.
            period_limits_path (string): The path to the period limits file.
            is_update (bool): Whether to update the period limits.

        Returns:
            dict: The response from the server.

        Examples:
        ```py
            response = client.upload_period_limits(1, 'path/to/my/file.csv')
            # {"message": "Information successfully added and uploaded to the request"}
        ```
        """
        with open(path, 'rb') as period_limits_file:
            try:
                if is_update:
                    return self.update_period_limits(
                        match_id,
                        files={
                            'file': (
                                'period_limits.csv',
                                period_limits_file,
                                'text/csv',
                            )
                        },
                    )
            except Exception as e:
                return e

            try:
                return self.post_period_limits(
                    match_id,
                    files={
                        'file': ('period_limits.csv', period_limits_file, 'text/csv')
                    },
                )
            except Exception as e:
                return e

    def upload_home_team_side(self, match_id: int, path: int, is_update: bool = False):
        """Post home team side to the server.

        Args:
            match_id (int): The match id.
            home_team_side_path (string): The path to the home team side file.
            is_update (bool): Whether to update the home team side.

        Returns:
            dict: The response from the server.

        Examples:
        ```py
            response = client.upload_home_team_side(1, 'path/to/my/file.csv')
            # {'message': 'Home Team Side successfully added and uploaded to the request'}
        ```
        """
        with open(path, 'rb') as home_team_side_file:
            try:
                if is_update:
                    return self.update_home_team_side(
                        match_id,
                        files={
                            'file': (
                                'home_team_side.csv',
                                home_team_side_file,
                                'text/csv',
                            )
                        },
                    )
            except Exception as e:
                return e

            try:
                return self.post_home_team_side(
                    match_id,
                    files={
                        'file': ('home_team_side.csv', home_team_side_file, 'text/csv')
                    },
                )
            except Exception as e:
                return e

    def get_match_physical(self, match_id: int, csv: bool = False):
        """Get the physical data of a match.

        Args:
            match_id (int): The match id.
            csv (bool): Whether to return the response as a CSV file.

        Returns:
            dict or file: The response from the server.

        Examples:
        ```py
            response = client.get_match_physical(match_id=1234556, csv=True)
            # CSV file or bytes response
        ```
        """
        try:
            physical_data = self.get_physical(match_id)
            if csv:
                with open(f'match_{match_id}_physical_data.csv', 'wb') as f:
                    f.write(physical_data)
                return f'match_{match_id}_physical_data.csv'
            return physical_data
        except Exception as e:
            return e

    def launch_match(self, match_id: int) -> dict:
        """Launch a match.

        Args:
            match_id (int): The match id.

        Returns:
            dict: The response from the server.s

        Examples:
        ```py
            response = client.launch_match(match_id=1)
            # {'message': 'Match added to processing queue'}
        ```
        """
        try:
            return self.post_launch_match(match_id)
        except Exception as e:
            return e

    def get_tracking(self, match_id: int):
        """Get the tracking data of a match and save it to a file.
        Warning: This method can take a long time to complete because of the large file size.
        Args:
            match_id (int): The match id.

        Returns:
            str: The name of the tracking data file.

        Examples:
        ```py
            response = client.get_tracking(match_id=1234567)
            # 'match_1234567_tracking_data.jsonl'
        ```
        """
        try:
            presigned_url = self.get_tracking_data(match_id)
            res = requests.get(presigned_url['url'])
            res.raise_for_status()
            with open(f'match_{match_id}_tracking_data.jsonl', 'wb') as f:
                f.write(res.content)
            return f'match_{match_id}_tracking_data.jsonl'
        except Exception as e:
            return e

    def get_data_collection(self, match_id: int):
        """Get the data collection of a match.
        Args:
            match_id (int): The match id.
        Returns:
            dict: The response from the server.
        """
        try:
            return self.data_collection(match_id)
        except Exception as e:
            return e

    def get_match_data(self, match_id: int):
        """Get the match data of a match.
        Args:
            match_id (int): The match id.
        Returns:
            dict: The response from the server.
        """
        try:
            return self.match_data(match_id)
        except Exception as e:
            return e

    def upload_match_list(self, path: str):
        """Post a match list to the server.
        Args:
            path (str): The path to the match list file.
        Returns:
            dict: The response from the server.
        """
        try:
            with open(path, 'rb') as f:
                print(Path(path).name)
                return self.post_match_list(
                    files={
                        'file': (
                            Path(path).name,
                            f,
                            'text/csv',
                        ),
                    }
                )
        except Exception as e:
            return e



from __future__ import annotations

import os
import re

DEFAULT_PATTERN = r'(?i)^(__version__|VERSION) *= *([\'"])v?(?P<version>.+?)\2'
DEFAULT_TEMPLATE = """\
# This file is auto-generated by Hatchling. As such, do not:
#   - modify
#   - track in version control e.g. be sure to add to .gitignore
__version__ = VERSION = {version!r}
"""


class VersionFile:
    def __init__(self, root: str, relative_path: str) -> None:
        self.__relative_path = relative_path
        self.__path = os.path.normpath(os.path.join(root, relative_path))
        self.__cached_read_data: tuple | None = None

    def read(self, *, pattern: str | bool) -> str:
        if not os.path.isfile(self.__path):
            message = f'file does not exist: {self.__relative_path}'
            raise OSError(message)

        with open(self.__path, encoding='utf-8') as f:
            contents = f.read()

        if not pattern or pattern is True:
            pattern = DEFAULT_PATTERN

        match = re.search(pattern, contents, flags=re.MULTILINE)
        if not match:
            message = f'unable to parse the version from the file: {self.__relative_path}'
            raise ValueError(message)

        groups = match.groupdict()
        if 'version' not in groups:
            message = 'no group named `version` was defined in the pattern'
            raise ValueError(message)

        self.__cached_read_data = groups['version'], contents, match.span('version')
        return self.__cached_read_data[0]

    def set_version(self, version: str) -> None:
        _old_version, file_contents, (start, end) = self.__cached_read_data  # type: ignore[misc]
        with open(self.__path, 'w', encoding='utf-8') as f:
            f.write(f'{file_contents[:start]}{version}{file_contents[end:]}')

    def write(self, version: str, template: str = DEFAULT_TEMPLATE) -> None:
        template = template or DEFAULT_TEMPLATE

        parent_dir = os.path.dirname(self.__path)
        if not os.path.isdir(parent_dir):
            os.makedirs(parent_dir)

        with open(self.__path, 'w', encoding='utf-8') as f:
            f.write(template.format(version=version))



from fitrequest.method_generator import RequestMethod

METHOD_DOCSTRING = (
    'Retrieve response from {endpoint} GET request. '
    'To learn more about it go to: https://tactical.skillcorner.com/api/docs/#{docs_url_anchor}.'
)


METHODS_BINDING = [
    {
        'name': 'get_all_requests',
        'endpoint': '/api/requests/',
        'docs_url_anchor': 'requests',
    },
    {
        'name': 'get_request',
        'endpoint': '/api/request/{}',
        'docs_url_anchor': 'request',
        'resource_name': 'match_id',
    },
    {
        'name': 'get_match_sheet',
        'endpoint': '/api/request/{}/match-sheet/',
        'docs_url_anchor': 'match_sheet',
        'resource_name': 'match_id',
    },
    {
        'name': 'post_match_sheet',
        'endpoint': '/api/request/{}/match-sheet/',
        'docs_url_anchor': 'match_sheet',
        'resource_name': 'match_id',
        'request_method': RequestMethod.post,
    },
    {
        'name': 'update_match_sheet',
        'endpoint': '/api/request/{}/match-sheet/',
        'docs_url_anchor': 'match_sheet',
        'resource_name': 'match_id',
        'request_method': RequestMethod.put,
    },
    {
        'name': 'get_period_limits',
        'endpoint': '/api/request/{}/period-limits/',
        'docs_url_anchor': 'period_limits',
        'resource_name': 'match_id',
    },
    {
        'name': 'post_period_limits',
        'endpoint': '/api/request/{}/period-limits/',
        'docs_url_anchor': 'period_limits',
        'resource_name': 'match_id',
        'request_method': RequestMethod.post,
    },
    {
        'name': 'update_period_limits',
        'endpoint': '/api/request/{}/period-limits/',
        'docs_url_anchor': 'period_limits',
        'resource_name': 'match_id',
        'request_method': RequestMethod.put,
    },
    {
        'name': 'get_home_team_side',
        'endpoint': '/api/request/{}/home-team-side/',
        'docs_url_anchor': 'home_team_side',
        'resource_name': 'match_id',
    },
    {
        'name': 'post_home_team_side',
        'endpoint': '/api/request/{}/home-team-side/',
        'docs_url_anchor': 'home_team_side',
        'resource_name': 'match_id',
        'request_method': RequestMethod.post,
    },
    {
        'name': 'update_home_team_side',
        'endpoint': '/api/request/{}/home-team-side/',
        'docs_url_anchor': 'home_team_side',
        'resource_name': 'match_id',
        'request_method': RequestMethod.put,
    },
    {
        'name': 'data_collection',
        'endpoint': '/api/match/{}/data_collection/',
        'docs_url_anchor': 'data_collection',
        'resource_name': 'match_id',
    },
    {
        'name': 'get_tracking_data',
        'endpoint': '/api/match/{}/tracking?data_version=3&on_demand=true',
        'docs_url_anchor': 'tracking',
        'resource_name': 'match_id',
    },
    {
        'name': 'match_data',
        'endpoint': '/api/match/{}/',
        'docs_url_anchor': 'match_data',
        'resource_name': 'match_id',
    },
    {
        'name': 'post_launch_match',
        'endpoint': '/api/request/{}/start-processing/',
        'docs_url_anchor': 'launch_match',
        'resource_name': 'match_id',
        'request_method': RequestMethod.post,
    },
    {
        'name': 'get_physical',
        'endpoint': '/api/match/{}/physical/',
        'docs_url_anchor': 'match_physical',
        'resource_name': 'match_id',
        'request_method': RequestMethod.get,
    },
    {
        'name': 'post_match_list',
        'endpoint': '/api/requests/match-list/',
        'docs_url_anchor': 'match_list',
        'request_method': RequestMethod.post,
    },
]
